/*
 * liftmon_snowcon_rpx-cllf v1.0 11/03/01
 * www.embeddedlinuxinterfacing.com
 *
 * The original location of this code is
 * http://www.embeddedlinuxinterfacing.com/chapters/09/
 *
 * Copyright (C) 2001 by Craig Hollabaugh
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*
powerpc-linux-gcc -O2 -D__KERNEL__ -DMODULE -I/usr/src/powerpc-linux/include \
-c liftmon_snowcon_rpxcllf.c \
-o /tftpboot/powerpc-rootfs/tmp/liftmon_snowcon_rpxcllf.o
*/

/*
 * liftmon_snowcon_mediaengine.c is based on procfs_example.c by Erik Mouw.
 * For more information, please see, The Linux Kernel Procfs Guide, Erik Mouw
 * http://kernelnewbies.org/documents/kdoc/procfs-guide/lkprocfsguide.html
 */

/* liftmon_snowcon
 * liftmon_snowcon uses read and write system calls to control
 * an interface circuit connected the RPX-CLLF's bus and expansion ports.
 * This module performs all bit operations for the data bus.
 * Bash script need only read and write to /proc entry files to determine
 * status or control equipment. In addition, the module's init code asserts
 * the OUTPUT_ENABLE signal and its exit code deasserts the OUTPUT_ENABLE
 * signal.
 *
 * This module creates these /proc entries:
 * Trailblazer directory  /proc/trailblazer
 * Lift Monitoring
 *  AC Mains              /proc/trailblazer/liftacmains
 *  Motor Controller      /proc/trailblazer/liftmotorcontroller
 *  Low Speed Operation   /proc/trailblazer/liftslowspeed
 *  High Speed Operation  /proc/trailblazer/lifthighspeed
 *  Operator Switch Base  /proc/trailblazer/liftoperatorswitchbase
 *  Operator Switch Top   /proc/trailblazer/liftoperatorswitchtop
 * Snow-Making Control
 *  Water Value 1         /proc/trailblazer/snowwatervalve1
 *  Water Value 2         /proc/trailblazer/snowwatervalve2
 *  Water Value 3         /proc/trailblazer/snowwatervalve3
 *  Heater                /proc/trailblazer/snowheater1
 */

/* references
 * see section 34.2 Port A MPC860 PowerQUICC User's Manual
 *     section 16.5 General-Purpose Chip-Select Machine (GPCM)
 */


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#include <asm/8xx_immap.h>


#define MODULE_VERSION "1.0"
#define MODULE_NAME "liftmon_snowcon_cllf"

volatile immap_t *immap;
static void *io_base;


#define EXPANSIONPORT          0x80000000
#define EXPANSIONPORTLEN       16

#define PA0                    0x8000


/* input bit definitions */
#define LIFTACMAINS            0x01
#define LIFTMOTORCONTROLLER    0x02
#define LIFTSLOWSPEED          0x04
#define LIFTHIGHSPEED          0x08
#define LIFTOPERATORSWITCHBASE 0x10
#define LIFTOPERATORSWITCHTOP  0x20

/* output bit definitions */
#define SNOWWATERVALVE1        0x01
#define SNOWWATERVALVE2        0x02
#define SNOWWATERVALVE3        0x04
#define SNOWHEATER1            0x08

/* define a bitmask, each *_file uses this to determine who it is */
struct liftmon_snowcon_data_t {
        unsigned char mask;
};

/* snowcondata is the output latch value stored internally. Control
 *  changes made by user scripts writing to /proc/trailblazer entries
 * result in bits being either cleared or set in snowcondata. We
 * write snowcondata to the output latch every time a control change
 * occurs
 */
unsigned char snowcondata;

struct liftmon_snowcon_data_t liftacmains_data,
                              liftmotorcontroller_data,
                              liftslowspeed_data,
                              lifthighspeed_data,
                              liftoperatorswitchbase_data,
                              liftoperatorswitchtop_data,
                              snowwatervalve1_data,
                              snowwatervalve2_data,
                              snowwatervalve3_data,
                              snowheater1_data;

static struct proc_dir_entry  *tb_dir,
                              *liftacmains_file,
                              *liftmotorcontroller_file,
                              *liftslowspeed_file,
                              *lifthighspeed_file,
                              *liftoperatorswitchbase_file,
                              *liftoperatorswitchtop_file,
                              *snowwatervalve1_file,
                              *snowwatervalve2_file,
                              *snowwatervalve3_file,
                              *snowheater1_file;

/* proc_read - proc_read_liftmon
 * proc_read_liftmon is the callback function that the kernel calls when
 * there's a read file operation on these /proc/trailblazer files:
 * liftacmains, lifthighspeed, liftmotorcontroller, liftoperatorswitchbase
 * liftoperatorswitchtop, and liftslowspeed. The file's data pointer is
 * passed in the data parameter. You first cast it to the
 * liftmon_snowcon_data_t structure. A read from io_base gets the AC input
 * status module. The result is anded with the bitmask value to
 * to determine if the particular input module is on or off. Which
 * particular input module is defined by which /proc/trailblazer/ file
 * is read.
 */
static int proc_read_liftmon(char *page, char **start, off_t off, int count,
                             int *eof, void *data)
{
  unsigned int v;
  struct liftmon_snowcon_data_t *liftmon_snowcon_data =
                                 (struct liftmon_snowcon_data_t *)data;

  v = readb(io_base);

/* mask the input value based on the mask. Each mask is different depending
 * which /proc/trailblazer file was read.
 * Electical note: returning an inverted value because AC power to an input
 * module pulls outputs a low and the input buffer, 74244, doesn't invert
 */
  if (v & liftmon_snowcon_data->mask)
    page[0] = '0';
  else
    page[0] = '1';

/* return 1 which is the length of page */
  return 1;
}

/* proc_write - proc_write_snowcondata
 * proc_write_snowcondata is the callback function that the kernel calls
 * when there's a write file operation on these /proc/trailblazer files:
 * snowheater1, snowwatervalve1, snowwatervalve2 and snowwatervalve3.
 * The file's data pointer is passed in the data parameter. You first
 * cast it to the liftmon_snowcon_data_t structure. The buffer parameter
 * points to the incoming data. If the incoming data is a 1 or a 0,
 * a bit in snowcondata is set or cleared. Which bit is defined by which
 * /proc/trailblazer file is written to. snowcondata is then written to
 * the output latch.
 */
static int proc_write_snowcondata(struct file *file, const char *buffer,
                                  unsigned long count, void *data)
{
  struct liftmon_snowcon_data_t *liftmon_snowcon_data = 
                                 (struct liftmon_snowcon_data_t *)data;

/* check if the user wrote a 1 or a 0 the /proc/trailblazer file.
   if so, set or clear a bit in snowcondata */
  if (buffer[0] == '0')
    snowcondata |= liftmon_snowcon_data->mask;

  if (buffer[0] == '1')
    snowcondata &= ~liftmon_snowcon_data->mask;

  writeb(snowcondata, io_base + 1);
  return 1;
}

/* proc_read - proc_read_snowcondata
 * proc_read_snowcondata is the callback function that the kernel calls
 * when there's a read file operation on these /proc/trailblazer files:
 * snowheater1, snowwatervalve1, snowwatervalve2 and snowwatervalve3.
 * The file's data pointer is passed in the data parameter. You first
 * cast it to the liftmon_snowcon_data_t structure.  Use snowcondata
 * anded with the bitmask value to determine if the particular output
 * module is on or off. Which particular output module is defined by
 * which /proc/trailblazer/ file is read.
 */
static int proc_read_snowcondata(char *page, char **start, off_t off, 
                                 int count, int *eof, void *data)
{
  struct liftmon_snowcon_data_t *liftmon_snowcon_data =
                                 (struct liftmon_snowcon_data_t *)data;

/* mask the snowcondata value based on the mask. Each mask is different
 * depending which /proc/trailblazer file was read. */
  if ( snowcondata & liftmon_snowcon_data->mask )
    page[0] = '0';
  else
    page[0] = '1';

/* return the length */
  return 1;
}

/* init - init_liftmon_snowcon
 * init_liftmon_snowcon creates the /proc entry files and obtains
 * their pointers. For each file, the fields, data, read_proc,
 * write_proc and owner, are filled.  Then the I/O controller is configured
 * so PA0 is an output and the memory controller works slowly with /CS5.
 * It initializes the output modules in the off state then
 * completes by writing an entry to the system log using printk.
 */
static int __init init_liftmon_snowcon(void)
{
  int rv = 0;

/* Create the trailblazer /proc entry */
  tb_dir = proc_mkdir("trailblazer", NULL);
  if(tb_dir == NULL) {
          rv = -ENOMEM;
          goto out;
  }
  tb_dir->owner = THIS_MODULE;

/* Create liftacmains and make it readable by all - 0444 */
  liftacmains_file = create_proc_entry("liftacmains", 0444, tb_dir);
  if(liftacmains_file == NULL) {
          rv = -ENOMEM;
          goto no_liftacmains;
  }
  liftacmains_data.mask = LIFTACMAINS;
  liftacmains_file->data = &liftacmains_data;
  liftacmains_file->read_proc = &proc_read_liftmon;
  liftacmains_file->write_proc = NULL;
  liftacmains_file->owner = THIS_MODULE;


/* Create liftmotorcontroller and make it readable by all - 0444 */
  liftmotorcontroller_file = create_proc_entry("liftmotorcontroller",
                                                0444, tb_dir);
  if(liftmotorcontroller_file == NULL) {
          rv = -ENOMEM;
          goto no_liftmotorcontroller;
  }
  liftmotorcontroller_data.mask = LIFTMOTORCONTROLLER;
  liftmotorcontroller_file->data = &liftmotorcontroller_data;
  liftmotorcontroller_file->read_proc = &proc_read_liftmon;
  liftmotorcontroller_file->write_proc = NULL;
  liftmotorcontroller_file->owner = THIS_MODULE;

/* Create liftslowspeed and make it readable by all - 0444 */
  liftslowspeed_file = create_proc_entry("liftslowspeed", 0444, tb_dir);
  if(liftslowspeed_file == NULL) {
          rv = -ENOMEM;
          goto no_liftslowspeed;
  }
  liftslowspeed_data.mask = LIFTSLOWSPEED;
  liftslowspeed_file->data = &liftslowspeed_data;
  liftslowspeed_file->read_proc = &proc_read_liftmon;
  liftslowspeed_file->write_proc = NULL;
  liftslowspeed_file->owner = THIS_MODULE;

/* Create lifthighspeed and make it readable by all - 0444 */
  lifthighspeed_file = create_proc_entry("lifthighspeed", 0444, tb_dir);
  if(lifthighspeed_file == NULL) {
          rv = -ENOMEM;
          goto no_lifthighspeed;
  }
  lifthighspeed_data.mask = LIFTHIGHSPEED;
  lifthighspeed_file->data = &lifthighspeed_data;
  lifthighspeed_file->read_proc = &proc_read_liftmon;
  lifthighspeed_file->write_proc = NULL;
  lifthighspeed_file->owner = THIS_MODULE;

/* Create liftoperatorswitchbase and make it readable by all - 0444 */
  liftoperatorswitchbase_file = create_proc_entry("liftoperatorswitchbase",
                                                   0444, tb_dir);
  if(liftoperatorswitchbase_file == NULL) {
          rv = -ENOMEM;
          goto no_liftoperatorswitchbase;
  }
  liftoperatorswitchbase_data.mask = LIFTOPERATORSWITCHBASE;
  liftoperatorswitchbase_file->data = &liftoperatorswitchbase_data;
  liftoperatorswitchbase_file->read_proc = &proc_read_liftmon;
  liftoperatorswitchbase_file->write_proc = NULL;
  liftoperatorswitchbase_file->owner = THIS_MODULE;

/* Create liftoperatorswitchtop and make it readable by all - 0444 */
  liftoperatorswitchtop_file = create_proc_entry("liftoperatorswitchtop",
                                                  0444, tb_dir);
  if(liftoperatorswitchtop_file == NULL) {
          rv = -ENOMEM;
          goto no_liftoperatorswitchtop;
  }
  liftoperatorswitchtop_data.mask = LIFTOPERATORSWITCHTOP;
  liftoperatorswitchtop_file->data = &liftoperatorswitchtop_data;
  liftoperatorswitchtop_file->read_proc = &proc_read_liftmon;
  liftoperatorswitchtop_file->write_proc = NULL;
  liftoperatorswitchtop_file->owner = THIS_MODULE;

/* Create snowwatervalve1 and make it root writeable, readable by all-0644 */
  snowwatervalve1_file = create_proc_entry("snowwatervalve1", 0644, tb_dir);
  if(snowwatervalve1_file == NULL) {
          rv = -ENOMEM;
          goto no_snowwatervalve1;
  }
  snowwatervalve1_data.mask = SNOWWATERVALVE1;
  snowwatervalve1_file->data = &snowwatervalve1_data;
  snowwatervalve1_file->read_proc = &proc_read_snowcondata;
  snowwatervalve1_file->write_proc = &proc_write_snowcondata;
  snowwatervalve1_file->owner = THIS_MODULE;

/* Create snowwatervalve2 and make it root writeable, readable by all-0644 */
  snowwatervalve2_file = create_proc_entry("snowwatervalve2", 0644, tb_dir);
  if(snowwatervalve2_file == NULL) {
          rv = -ENOMEM;
          goto no_snowwatervalve2;
  }
  snowwatervalve2_data.mask = SNOWWATERVALVE2;
  snowwatervalve2_file->data = &snowwatervalve2_data;
  snowwatervalve2_file->read_proc = &proc_read_snowcondata;
  snowwatervalve2_file->write_proc = &proc_write_snowcondata;
  snowwatervalve2_file->owner = THIS_MODULE;

/* Create snowwatervalve3 and make it root writeable, readable by all-0644 */
  snowwatervalve3_file = create_proc_entry("snowwatervalve3", 0644, tb_dir);
  if(snowwatervalve3_file == NULL) {
          rv = -ENOMEM;
          goto no_snowwatervalve3;
  }
  snowwatervalve3_data.mask = SNOWWATERVALVE3;
  snowwatervalve3_file->data = &snowwatervalve3_data;
  snowwatervalve3_file->read_proc = &proc_read_snowcondata;
  snowwatervalve3_file->write_proc = &proc_write_snowcondata;
  snowwatervalve3_file->owner = THIS_MODULE;

/* Create snowheater1 and make it root writeable, readable by all-0644 */
  snowheater1_file = create_proc_entry("snowheater1", 0644, tb_dir);
  if(snowheater1_file == NULL) {
          rv = -ENOMEM;
          goto no_snowheater1;
  }
  snowheater1_data.mask = SNOWHEATER1;
  snowheater1_file->data = &snowheater1_data;
  snowheater1_file->read_proc = &proc_read_snowcondata;
  snowheater1_file->write_proc = &proc_write_snowcondata;
  snowheater1_file->owner = THIS_MODULE;


/* get the register base address */
  immap = (immap_t *)(mfspr(IMMR) & 0xFFFF0000);

/* set CS5 for memory at 0x80000000, see 16.4.1 Base Registers (BRx)
 * and 16.4.2 Option Registers (ORx) for slow operation
 */
  immap->im_memctl.memc_br5 = 0x80000401;
  immap->im_memctl.memc_or5 = 0x80000FF6;

/* get the remapped GPIO controller base address register */
  io_base = ioremap_nocache(EXPANSIONPORT, EXPANSIONPORTLEN);

/* initialize snowcondata to 0xFF, all output modules off */
/* output modules invert */
  snowcondata = 0xFF;
  writeb(snowcondata, io_base);

  immap->im_ioport.iop_papar &= ~PA0; /* setting PA0 to general I/O */
  immap->im_ioport.iop_padir |=  PA0; /* make PA0 an output         */
  immap->im_ioport.iop_padat &= ~PA0; /* enable the 74HC574 output  */

  printk("immr    = 0x%08X\n",immap);
  printk("io_base = 0x%08X\n", io_base);
  printk("BR5     = 0x%08X\n",immap->im_memctl.memc_br5);
  printk("OR5     = 0x%08X\n",immap->im_memctl.memc_or5);
  printk("PAPAR   = 0x%04X\n",immap->im_ioport.iop_papar);
  printk("PADIR   = 0x%04X\n",immap->im_ioport.iop_padir);

/* everything initialized */
  printk(KERN_INFO "%s %s initialized\n",MODULE_NAME, MODULE_VERSION);
  return 0;

/* this removes /proc entries if we have an error along the way */
no_snowheater1:
  remove_proc_entry("snowheater1", tb_dir);
no_snowwatervalve3:
  remove_proc_entry("snowwatervalve3", tb_dir);
no_snowwatervalve2:
  remove_proc_entry("snowwatervalve2", tb_dir);
no_snowwatervalve1:
  remove_proc_entry("snowwatervalve1", tb_dir);
no_liftoperatorswitchtop:
  remove_proc_entry("liftoperatorswitchtop", tb_dir);
no_liftoperatorswitchbase:
  remove_proc_entry("liftoperatorswitchbase", tb_dir);
no_lifthighspeed:
  remove_proc_entry("lifthighspeed", tb_dir);
no_liftslowspeed:
  remove_proc_entry("liftslowspeed", tb_dir);
no_liftmotorcontroller:
  remove_proc_entry("liftmotorcontroller", tb_dir);
no_liftacmains:
  remove_proc_entry("liftacmains", tb_dir);
out:
  return rv;
}

/* exit - cleanup_liftmon_snowcon
 * cleanup_liftmon_snowcon deasserts the OUTPUT_ENABLE signal.
 * It removes the /proc entry files prints a message to the system log.
 */
static void __exit cleanup_liftmon_snowcon(void)
{
  immap->im_ioport.iop_padir &= ~PA0; /* make PA0 an input */

/* release the ioremap */
  iounmap(io_base);

/* removing the /proc entries */
  remove_proc_entry("liftacmains", tb_dir);
  remove_proc_entry("liftmotorcontroller", tb_dir);
  remove_proc_entry("liftslowspeed", tb_dir);
  remove_proc_entry("lifthighspeed", tb_dir);
  remove_proc_entry("liftoperatorswitchbase", tb_dir);
  remove_proc_entry("liftoperatorswitchtop", tb_dir);
  remove_proc_entry("snowwatervalve1", tb_dir);
  remove_proc_entry("snowwatervalve2", tb_dir);
  remove_proc_entry("snowwatervalve3", tb_dir);
  remove_proc_entry("snowheater1", tb_dir);
  remove_proc_entry("trailblazer", NULL);

  printk(KERN_INFO "%s %s removed\n", MODULE_NAME, MODULE_VERSION);
}

module_init(init_liftmon_snowcon);
module_exit(cleanup_liftmon_snowcon);

MODULE_AUTHOR("Craig Hollabaugh");
MODULE_DESCRIPTION(
"Trailblazer Lift Monitor and Snowmaking Control for RPX-CLLF");

EXPORT_NO_SYMBOLS;
